from typing import NamedTuple
import io
from PIL import Image


class Image32(NamedTuple):
    data: bytes
    width: int
    height: int

    @staticmethod
    def load(src: bytes) -> 'Image32':
        image = Image.open(io.BytesIO(src))
        if image.mode != 'RGBA':
            image = image.convert('RGBA')        
        data = image.tobytes('raw')
        return Image32(data, image.width, image.height)


class Texture:
    def __init__(self, name: str, image: Image32) -> None:
        self.name = name
        self.image = image



from OpenGL import GL
from OpenGL.raw.GL.VERSION.GL_1_0 import glTexImage2D


class Texture:
    def __init__(self) -> None:
        self.texture = GL.glGenTextures(1)

    def __del__(self) -> None:
        GL.glDeleteTextures(1, [self.texture])

    def bind(self):
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.texture)

    def unbind(self):
        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)

    def set_image(self, data: bytes, width: int, height: int):
        self.bind()
        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_RGBA, width,
                        height, 0, GL.GL_RGBA, GL.GL_UNSIGNED_BYTE, data)
        GL.glTexParameterf(
            GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR)
        GL.glTexParameterf(
            GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR)
        self.unbind()

    def activate(self, location: int, slot: int):
        self.bind()
        GL.glActiveTexture(GL.GL_TEXTURE0 + slot)  # type: ignore
        GL.glUniform1i(location, slot)



import ctypes
from OpenGL import GL
from .mesh import Mesh
from .material import Material
from .vertices import Planar, TypedBytes
import pkgutil


class Float3(ctypes.Structure):
    _fields_ = [
        ('x', ctypes.c_float),
        ('y', ctypes.c_float),
        ('z', ctypes.c_float),
    ]


def create_axis(size: float) -> Mesh:
    positions = (Float3 * 12)(
        Float3(0,
               0,
               0),
        Float3(size,
               0,
               0),
        Float3(0,
               0,
               0),
        Float3(-size,
               0,
               0),
        Float3(0,
               0,
               0),
        Float3(0,
               size,
               0),
        Float3(0,
               0,
               0),
        Float3(0,
               -size,
               0),
        Float3(0,
               0,
               0),
        Float3(0,
               0,
               size),
        Float3(0,
               0,
               0),
        Float3(0,
               0,
               -size),
    )
    colors = (Float3 * 12)(
        Float3(1,
               0,
               0),
        Float3(1,
               0,
               0),
        Float3(0.5,
               0,
               0),
        Float3(0.5,
               0,
               0),
        Float3(0,
               1,
               0),
        Float3(0,
               1,
               0),
        Float3(0,
               0.5,
               0),
        Float3(0,
               0.5,
               0),
        Float3(0,
               0,
               1),
        Float3(0,
               0,
               1),
        Float3(0,
               0,
               0.5),
        Float3(0,
               0,
               0.5),
    )
    mesh = Mesh(f'axis {size}', Planar([
        TypedBytes(memoryview(
            positions).tobytes(), ctypes.c_float, 3),
        TypedBytes(memoryview(
            colors).tobytes(), ctypes.c_float, 3),
    ]))
    vs = pkgutil.get_data('glglue', 'assets/axis.vs')
    if not vs:
        raise Exception()
    fs = pkgutil.get_data('glglue', 'assets/axis.fs')
    if not fs:
        raise Exception()
    material = Material('axis', vs.decode('utf-8'), fs.decode('utf-8'))
    mesh.add_submesh(material, [], GL.GL_LINES)
    return mesh



from logging import getLogger
from typing import Any, List, NamedTuple
from OpenGL import GL
import glglue.basecontroller
import glglue.ctypesmath
import glglue.gl3.vbo
import glglue.scene.material
from ..scene import cube, axis
from . renderer import Renderer
logger = getLogger(__name__)


class SampleController(glglue.basecontroller.BaseController):
    def __init__(self):
        self.clear_color = (0.6, 0.6, 0.4, 0.0)
        self.axis = axis.create_axis(1.0)
        self.drawables: List[Any] = [cube.create_cube(0.3)]
        self.camera = glglue.ctypesmath.Camera()
        self.isInitialized = False
        self.renderer = Renderer()

    def onResize(self, w: int, h: int) -> bool:
        GL.glViewport(0, 0, w, h)
        return self.camera.onResize(w, h)

    def onLeftDown(self, x: int, y: int) -> bool:
        ''' mouse input '''
        return self.camera.onLeftDown(x, y)

    def onLeftUp(self, x: int, y: int) -> bool:
        ''' mouse input '''
        return self.camera.onLeftUp(x, y)

    def onMiddleDown(self, x: int, y: int) -> bool:
        ''' mouse input '''
        return self.camera.onMiddleDown(x, y)

    def onMiddleUp(self, x: int, y: int) -> bool:
        ''' mouse input '''
        return self.camera.onMiddleUp(x, y)

    def onRightDown(self, x: int, y: int) -> bool:
        ''' mouse input '''
        return self.camera.onRightDown(x, y)

    def onRightUp(self, x: int, y: int) -> bool:
        ''' mouse input '''
        return self.camera.onRightUp(x, y)

    def onMotion(self, x: int, y: int) -> bool:
        ''' mouse input '''
        return self.camera.onMotion(x, y)

    def onWheel(self, d: int) -> bool:
        ''' mouse input '''
        return self.camera.onWheel(d)

    def onKeyDown(self, keycode: int) -> bool:
        return False

    def onUpdate(self, d: int) -> bool:
        '''
        milliseconds
        '''
        for drawable in self.drawables:
            drawable.update(d)
        return False

    def initialize(self):
        GL.glEnable(GL.GL_DEPTH_TEST)
        self.isInitialized = True

    def draw(self):
        if not self.isInitialized:
            self.initialize()
        GL.glClearColor(*self.clear_color)
        GL.glClear(GL.GL_COLOR_BUFFER_BIT |
                   GL.GL_DEPTH_BUFFER_BIT)  # type: ignore

        self.renderer.draw(
            self.axis, self.camera.projection.matrix, self.camera.view.matrix)
        for drawable in self.drawables:
            self.renderer.draw(drawable, self.camera.projection.matrix,
                               self.camera.view.matrix)

        GL.glFlush()



from typing import NamedTuple, Type, List
import ctypes
import array


class TypedBytes(NamedTuple):
    data: bytes
    element_type: Type[ctypes._SimpleCData]
    element_count: int = 1

    @staticmethod
    def create(src: array.array) -> 'TypedBytes':
        match src:
            case array.array() as a:
                match a.typecode:
                    case 'H':
                        return TypedBytes(memoryview(a).tobytes(), ctypes.c_ushort)
        raise RuntimeError()

    def stride(self) -> int:
        return ctypes.sizeof(self.element_type) * self.element_count

    def count(self) -> int:
        return len(self.data) // self.stride()


class Planar(NamedTuple):
    attributes: List[TypedBytes]

    def count(self) -> int:
        return self.attributes[0].count()


class Interleaved(NamedTuple):
    vertices: TypedBytes
    offsets: List[int]

    def count(self) -> int:
        return self.vertices.count()



import ctypes
import math
from typing import Tuple


class Mat4(ctypes.Structure):
    _fields_ = [("_11", ctypes.c_float), ("_12", ctypes.c_float),
                ("_13", ctypes.c_float), ("_14", ctypes.c_float),
                ("_21", ctypes.c_float), ("_22", ctypes.c_float),
                ("_23", ctypes.c_float), ("_24", ctypes.c_float),
                ("_31", ctypes.c_float), ("_32", ctypes.c_float),
                ("_33", ctypes.c_float), ("_34", ctypes.c_float),
                ("_41", ctypes.c_float), ("_42", ctypes.c_float),
                ("_43", ctypes.c_float), ("_44", ctypes.c_float)]

    def __str__(self) -> str:
        return f'[{self._11}, {self._12}, {self._13}, {self._14}]' + f'[{self._21}, {self._22}, {self._23}, {self._24}]' + f'[{self._31}, {self._32}, {self._33}, {self._34}]' + f'[{self._41}, {self._42}, {self._43}, {self._44}]'

    def __mul__(self, rhs: 'Mat4') -> 'Mat4':
        m = Mat4()
        m._11 = self._11 * rhs._11 + self._12 * rhs._21 + \
            self._13 * rhs._31 + self._14 * rhs._41
        m._12 = self._11 * rhs._12 + self._12 * rhs._22 + \
            self._13 * rhs._32 + self._14 * rhs._42
        m._13 = self._11 * rhs._13 + self._12 * rhs._23 + \
            self._13 * rhs._33 + self._14 * rhs._43
        m._14 = self._11 * rhs._14 + self._12 * rhs._24 + \
            self._13 * rhs._34 + self._14 * rhs._44

        m._21 = self._21 * rhs._11 + self._22 * rhs._21 + \
            self._23 * rhs._31 + self._24 * rhs._41
        m._22 = self._21 * rhs._12 + self._22 * rhs._22 + \
            self._23 * rhs._32 + self._24 * rhs._42
        m._23 = self._21 * rhs._13 + self._22 * rhs._23 + \
            self._23 * rhs._33 + self._24 * rhs._43
        m._24 = self._21 * rhs._14 + self._22 * rhs._24 + \
            self._23 * rhs._34 + self._24 * rhs._44

        m._31 = self._31 * rhs._11 + self._32 * rhs._21 + \
            self._33 * rhs._31 + self._34 * rhs._41
        m._32 = self._31 * rhs._12 + self._32 * rhs._22 + \
            self._33 * rhs._32 + self._34 * rhs._42
        m._33 = self._31 * rhs._13 + self._32 * rhs._23 + \
            self._33 * rhs._33 + self._34 * rhs._43
        m._34 = self._31 * rhs._14 + self._32 * rhs._24 + \
            self._33 * rhs._34 + self._34 * rhs._44

        m._41 = self._41 * rhs._11 + self._42 * rhs._21 + \
            self._43 * rhs._31 + self._44 * rhs._41
        m._42 = self._41 * rhs._12 + self._42 * rhs._22 + \
            self._43 * rhs._32 + self._44 * rhs._42
        m._43 = self._41 * rhs._13 + self._42 * rhs._23 + \
            self._43 * rhs._33 + self._44 * rhs._43
        m._44 = self._41 * rhs._14 + self._42 * rhs._24 + \
            self._43 * rhs._34 + self._44 * rhs._44
        return m

    def apply(self, x: float, y: float, z: float) -> Tuple[float, float, float]:
        return (
            x * self._11 + y * self._21 + z * self._31 + self._41,
            x * self._12 + y * self._22 + z * self._32 + self._42,
            x * self._13 + y * self._23 + z * self._33 + self._43
        )

    def to_array(self):
        return (ctypes.c_float * 16).from_buffer(self)

    def perspective(self, fov_y: float, aspect: float, z_near: float,
                    z_far: float) -> None:
        fov_y *= 0.5
        cot = 1.0 / math.tan(fov_y)
        self._11 = cot / aspect
        self._12 = 0
        self._13 = 0
        self._14 = 0
        self._21 = 0
        self._22 = cot
        self._23 = 0
        self._24 = 0
        self._31 = 0
        self._32 = 0
        self._33 = -(z_far + z_near) / (z_far - z_near)
        self._34 = -1
        self._41 = 0
        self._42 = 0
        self._43 = -2 * z_far * z_near / (z_far - z_near)
        self._44 = 0

    @classmethod
    def new_perspective(cls, fov_y, aspect, z_near, z_far) -> 'Mat4':
        m = cls()
        m.perspective(fov_y, aspect, z_near, z_far)
        return m

    @classmethod
    def new_identity(cls):
        return cls(
            1.0,
            0.0,
            0.0,
            0.0,  #
            0.0,
            1.0,
            0.0,
            0.0,  #
            0.0,
            0.0,
            1.0,
            0.0,  #
            0.0,
            0.0,
            0.0,
            1.0  #
        )

    @classmethod
    def new_translation(cls, x, y, z):
        return cls(
            1, 0, 0, 0,  #
            0, 1, 0, 0,  #
            0, 0, 1, 0,  #
            x, y, z, 1  #
        )

    @classmethod
    def new_rotation_z(cls, rad):
        s = math.sin(rad)
        c = math.cos(rad)
        return cls(
            c, s, 0, 0,  #
            -s, c, 0, 0,  #
            0, 0, 1, 0,  #
            0, 0, 0, 1  #
        )

    @classmethod
    def new_rotation_y(cls, rad):
        s = math.sin(rad)
        c = math.cos(rad)
        return cls(
            c, 0, -s, 0,  #
            0, 1, 0, 0,  #
            s, 0, c, 0,  #
            0, 0, 0, 1  #
        )

    @classmethod
    def new_rotation_x(cls, rad):
        s = math.sin(rad)
        c = math.cos(rad)
        return cls(
            1, 0, 0, 0,  #
            0, c, s, 0,  #
            0, -s, c, 0,  #
            0, 0, 0, 1  #
        )


class Vec4(ctypes.Structure):
    _fields_ = [("x", ctypes.c_float), ("y", ctypes.c_float),
                ("z", ctypes.c_float), ("w", ctypes.c_float)]

    def __mul__(self, m: Mat4) -> None:
        v = Vec4()
        v.x = self.x * m._11 + self.y * m._21 + self.z * m._31 + self.w * m._41
        v.y = self.x * m._12 + self.y * m._22 + self.z * m._32 + self.w * m._42
        v.z = self.x * m._13 + self.y * m._23 + self.z * m._33 + self.w * m._43
        v.w = self.x * m._14 + self.y * m._24 + self.z * m._34 + self.w * m._44
        return v



from typing import List
from ..ctypesmath.mat4 import Mat4
from . import mesh


class Node:
    def __init__(self, name: str):
        self. name = name
        self.model_matrix = Mat4.new_identity()
        self.children: List['Node'] = []
        self.meshes: List[mesh.Mesh] = []

    def update(self, delta):
        pass



from logging import getLogger
logger = getLogger(__name__)

import sys
sys.path.append('..')
sys.path.append('.')
from ctypes import *
from ctypes.wintypes import *

from OpenGL.GL import *
from glglue import win32con


##############################################################################
# Windows types
##############################################################################
def ErrorIfZero(handle):
    if handle == 0:
        raise WinError()
    else:
        return handle


def LOWORD(n):
    return n & 0xFFFF


def HIWORD(n):
    return (n >> 16) & 0XFFFF


WNDPROC = WINFUNCTYPE(c_int, HWND, c_uint, WPARAM, LPARAM)

win32con.PFD_TYPE_RGBA = 0
win32con.PFD_TYPE_COLORINDEX = 1
win32con.PFD_MAIN_PLANE = 0
win32con.PFD_OVERLAY_PLANE = 1
win32con.PFD_UNDERLAY_PLANE = (-1)
win32con.PFD_DOUBLEBUFFER = 0x00000001
win32con.PFD_STEREO = 0x00000002
win32con.PFD_DRAW_TO_WINDOW = 0x00000004
win32con.PFD_DRAW_TO_BITMAP = 0x00000008
win32con.PFD_SUPPORT_GDI = 0x00000010
win32con.PFD_SUPPORT_OPENGL = 0x00000020
win32con.PFD_GENERIC_FORMAT = 0x00000040
win32con.PFD_NEED_PALETTE = 0x00000080
win32con.PFD_NEED_SYSTEM_PALETTE = 0x00000100
win32con.PFD_SWAP_EXCHANGE = 0x00000200
win32con.PFD_SWAP_COPY = 0x00000400
win32con.PFD_SWAP_LAYER_BUFFERS = 0x00000800
win32con.PFD_GENERIC_ACCELERATED = 0x00001000
win32con.PFD_DEPTH_DONTCARE = 0x20000000
win32con.PFD_DOUBLEBUFFER_DONTCARE = 0x40000000
win32con.PFD_STEREO_DONTCARE = 0x80000000

WGL_ARB_create_context = 1
WGL_CONTEXT_DEBUG_BIT_ARB = 0x00000001
WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB = 0x00000002
WGL_CONTEXT_MAJOR_VERSION_ARB = 0x2091
WGL_CONTEXT_MINOR_VERSION_ARB = 0x2092
WGL_CONTEXT_LAYER_PLANE_ARB = 0x2093
WGL_CONTEXT_FLAGS_ARB = 0x2094
WGL_CONTEXT_PROFILE_MASK_ARB = 0x9126
WGL_CONTEXT_CORE_PROFILE_BIT_ARB = 0x00000001
WGL_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB = 0x00000002


class WNDCLASS(Structure):
    _fields_ = [('style', c_uint), ('lpfnWndProc', WNDPROC),
                ('cbClsExtra', c_int), ('cbWndExtra', c_int),
                ('hInstance', HANDLE), ('hIcon', HANDLE), ('hCursor', HANDLE),
                ('hbrBackground', HANDLE), ('lpszMenuName', LPCSTR),
                ('lpszClassName', LPCSTR)]


class CREATESTRUCT(Structure):
    _fields_ = [('lpCreateParams', c_void_p), ('hInstance', HANDLE),
                ('hMenu', HANDLE), ('hwndParent', HWND), ('cy', c_int),
                ('cx', c_int), ('y', c_int), ('x', c_int), ('style', c_int),
                ('lpszName', LPCSTR), ('lpszClass', LPCSTR),
                ('dwExStyle', c_int)]


class PIXELFORMATDESCRIPTOR(Structure):
    _fields_ = [
        ('nSize', c_ushort),
        ('nVersion', c_ushort),
        ('dwFlags', c_ulong),
        ('iPixelType', c_ubyte),
        ('cColorBits', c_ubyte),
        ('cRedBits', c_ubyte),
        ('cRedShift', c_ubyte),
        ('cGreenBits', c_ubyte),
        ('cGreenShift', c_ubyte),
        ('cBlueBits', c_ubyte),
        ('cBlueShift', c_ubyte),
        ('cAlphaBits', c_ubyte),
        ('cAlphaShift', c_ubyte),
        ('cAccumBits', c_ubyte),
        ('cAccumRedBits', c_ubyte),
        ('cAccumGreenBits', c_ubyte),
        ('cAccumBlueBits', c_ubyte),
        ('cAccumAlphaBits', c_ubyte),
        ('cDepthBits', c_ubyte),
        ('cStencilBits', c_ubyte),
        ('cAuxBuffers', c_ubyte),
        ('iLayerType', c_ubyte),
        ('bReserved', c_ubyte),
        ('dwLayerMask', c_ulong),
        ('dwVisibleMask', c_ulong),
        ('dwDamageMask', c_ulong),
    ]


class RECT(Structure):
    _fields_ = [('left', c_long), ('top', c_long), ('right', c_long),
                ('bottom', c_long)]


class PAINTSTRUCT(Structure):
    _fields_ = [('hdc', HDC), ('fErase', BOOL), ('rcPaint', RECT),
                ('fRestore', BOOL), ('fIncUpdate', BOOL),
                ('rgbReserved', c_char * 32)]


class POINT(Structure):
    _fields_ = [('x', c_long), ('y', c_long)]


class MSG(Structure):
    _fields_ = [('hwnd', HWND), ('message', c_uint), ('wParam', WPARAM),
                ('lParam', LPARAM), ('time', c_int), ('pt', POINT)]


ShowWindow = windll.user32.ShowWindow
ShowWindow.argtypes = [HWND, c_int]
ShowWindow.restype = BOOL

RegisterClass = windll.user32.RegisterClassA
RegisterClass.argtypes = [c_void_p]
RegisterClass.restype = ATOM

CreateWindowEx = windll.user32.CreateWindowExA
CreateWindowEx.argtypes = [
    c_int, LPCSTR, LPCSTR, c_int, c_int, c_int, c_int, c_int, HWND, HANDLE,
    HANDLE, py_object
]
CreateWindowEx.restype = HWND

if sys.maxsize > 2**32:
    # 64bit
    SetWindowLongPtr = windll.user32.SetWindowLongPtrA
    SetWindowLongPtr.argtypes = [HWND, c_int, c_void_p]
    SetWindowLongPtr.restype = c_void_p

    GetWindowLongPtr = windll.user32.GetWindowLongPtrA
    GetWindowLongPtr.argtypes = [HWND, c_int]
    GetWindowLongPtr.restype = c_void_p
else:
    # 64bit
    SetWindowLongPtr = windll.user32.SetWindowLongA
    SetWindowLongPtr.argtypes = [HWND, c_int, c_void_p]
    SetWindowLongPtr.restype = c_void_p

    GetWindowLongPtr = windll.user32.GetWindowLongA
    GetWindowLongPtr.argtypes = [HWND, c_int]
    GetWindowLongPtr.restype = c_void_p

GetDC = windll.user32.GetDC
GetDC.argtypes = [HWND]

BeginPaint = windll.user32.BeginPaint
BeginPaint.argtypes = [HWND, c_void_p]
BeginPaint.restype = HDC

EndPaint = windll.user32.EndPaint
EndPaint.argtypes = [HWND, c_void_p]
EndPaint.restype = HDC

ChoosePixelFormat = windll.gdi32.ChoosePixelFormat
ChoosePixelFormat.argtypes = [HDC, c_void_p]
ChoosePixelFormat.restype = ErrorIfZero

SetPixelFormat = windll.gdi32.SetPixelFormat
SetPixelFormat.argtypes = [
    HDC,
    c_int,
    c_void_p,
]
SetPixelFormat.restype = ErrorIfZero

SwapBuffers = windll.gdi32.SwapBuffers
SwapBuffers.argtypes = [
    HDC,
]
SwapBuffers.restype = ErrorIfZero

wglCreateContext = windll.opengl32.wglCreateContext
wglCreateContext.argtypes = [
    HDC,
]
wglCreateContext.restype = ErrorIfZero

wglDeleteContext = windll.opengl32.wglDeleteContext

wglMakeCurrent = windll.opengl32.wglMakeCurrent
wglMakeCurrent.argtypes = [
    HDC,
    HANDLE,
]
wglMakeCurrent.restype = ErrorIfZero

wglGetProcAddress = windll.opengl32.wglGetProcAddress
wglGetProcAddress.argtypes = [c_void_p]
wglGetProcAddress.restype = ErrorIfZero

DefWindowProc = windll.user32.DefWindowProcA
DefWindowProc.argtypes = [HWND, c_int, WPARAM, LPARAM]
DefWindowProc.restype = c_int

timeGetTime = windll.winmm.timeGetTime
timeGetTime.argtypes = []
timeGetTime.restype = c_ulong

##############################################################################
# api class
##############################################################################


class Window(object):
    def __init__(self):
        self.hwnd = None
        self.hrc = None
        self.controller = None
        self.ptrs = {}
        self.callbacks = {
            win32con.WM_PAINT: self.onPaint,
            win32con.WM_SIZE: self.onSize,
            win32con.WM_ERASEBKGND: self.onErasebkgnd,
            win32con.WM_DESTROY: self.onDestroy,
            win32con.WM_LBUTTONDOWN: self.onLeftDown,
            win32con.WM_LBUTTONUP: self.onLeftUp,
            win32con.WM_MBUTTONDOWN: self.onMiddleDown,
            win32con.WM_MBUTTONUP: self.onMiddleUp,
            win32con.WM_RBUTTONDOWN: self.onRightDown,
            win32con.WM_RBUTTONUP: self.onRightUp,
            win32con.WM_MOUSEMOVE: self.onMouseMove,
            win32con.WM_MOUSEWHEEL: self.onMouseWheel,
            win32con.WM_KEYDOWN: self.onKeyDown,
        }

    def finalize(self):
        logger.debug('finalize %s', self.__class__)

    def Redraw(self):
        windll.user32.InvalidateRect(self.hwnd, c_int(0), c_int(0))

    def Show(self):
        ShowWindow(self.hwnd, win32con.SW_SHOWNORMAL)
        windll.user32.UpdateWindow(self.hwnd)

    def SetLongPtr(self, hwnd, key, value):
        self.ptrs[key] = value
        SetWindowLongPtr(hwnd, key, value)

    def onPaint(self, hwnd, message, wParam, lParam):
        ps = PAINTSTRUCT()
        hdc = BeginPaint(hwnd, byref(ps))
        # if self.hrc and self.controller:
        #     self.controller.draw()
        #     SwapBuffers(hdc)
        EndPaint(hwnd, byref(ps))
        return 0

    def onSize(self, hwnd, message, wParam, lParam):
        w = LOWORD(lParam)
        h = HIWORD(lParam)
        # logger.debug("WM_SIZE %d x %d %s" % (w, h, self.controller))
        if self.controller:
            self.controller.onResize(w, h)
        return 0

    def onErasebkgnd(self, hwnd, message, wParam, lParam):
        return 0

    def onDestroy(self, hwnd, message, wParam, lParam):
        # sys.stderr.write("WM_DESTROY\n")
        windll.user32.PostQuitMessage(0)
        return 0

    def onKeyDown(self, hwnd, message, wParam, lParam):
        if self.controller:
            if wParam >= 65 and wParam <= 90:
                # lower
                if self.controller.onKeyDown(wParam + 32):
                    self.Redraw()
            else:
                if self.controller.onKeyDown(wParam):
                    self.Redraw()
        return 0

    def onMouseMove(self, hwnd, message, wParam, lParam):
        if self.controller:
            x = LOWORD(lParam)
            y = HIWORD(lParam)
            if self.controller.onMotion(x, y):
                self.Redraw()
        return 0

    def onMouseWheel(self, hwnd, message, wParam, lParam):
        if self.controller:
            d = HIWORD(wParam)
            if self.controller.onWheel(d > 32767 and 65536 - d or -d):
                self.Redraw()
        return 0

    def onLeftUp(self, hwnd, message, wParam, lParam):
        if self.controller:
            x = LOWORD(lParam)
            y = HIWORD(lParam)
            if self.controller.onLeftUp(x, y):
                self.Redraw()
        return 0

    def onLeftDown(self, hwnd, message, wParam, lParam):
        if self.controller:
            x = LOWORD(lParam)
            y = HIWORD(lParam)
            if self.controller.onLeftDown(x, y):
                self.Redraw()
        return 0

    def onMiddleUp(self, hwnd, message, wParam, lParam):
        if self.controller:
            x = LOWORD(lParam)
            y = HIWORD(lParam)
            if self.controller.onMiddleUp(x, y):
                self.Redraw()
        return 0

    def onMiddleDown(self, hwnd, message, wParam, lParam):
        if self.controller:
            x = LOWORD(lParam)
            y = HIWORD(lParam)
            if self.controller.onMiddleDown(x, y):
                self.Redraw()
        return 0

    def onRightUp(self, hwnd, message, wParam, lParam):
        if self.controller:
            x = LOWORD(lParam)
            y = HIWORD(lParam)
            if self.controller.onRightUp(x, y):
                self.Redraw()
        return 0

    def onRightDown(self, hwnd, message, wParam, lParam):
        if self.controller:
            x = LOWORD(lParam)
            y = HIWORD(lParam)
            if self.controller.onRightDown(x, y):
                self.Redraw()
        return 0

    def WndProc(self, hwnd, message, wParam, lParam):
        if message in self.callbacks:
            return self.callbacks[message](hwnd, message, wParam, lParam)
        else:
            return DefWindowProc(hwnd, message, wParam, lParam)

    def createGLContext(self, bits, version=None, profile='compatibility'):
        hdc = GetDC(self.hwnd)

        pfd = PIXELFORMATDESCRIPTOR()
        pfd.nsize = sizeof(PIXELFORMATDESCRIPTOR)
        pfd.nVersion = 1
        pfd.dwFlags = (win32con.PFD_DRAW_TO_WINDOW
                       | win32con.PFD_SUPPORT_OPENGL
                       | win32con.PFD_DOUBLEBUFFER)
        pfd.iPixelType = win32con.PFD_TYPE_RGBA
        pfd.cColorBits = bits
        pfd.cRedBits = 0
        pfd.cRedShift = 0
        pfd.cGreenBits = 0
        pfd.cGreenShift = 0
        pfd.cBlueBits = 0
        pfd.cBlueShift = 0
        pfd.cAlphaBits = 0
        pfd.cAlphaShift = 0
        pfd.cAccumBits = 0
        pfd.cAccumRedBits = 0
        pfd.cAccumGreenBits = 0
        pfd.cAccumBlueBits = 0
        pfd.cAccumAlphaBits = 0
        pfd.cDepthBits = bits
        pfd.cStencilBits = 2
        pfd.cAuxBuffers = 0
        pfd.iLayerType = win32con.PFD_MAIN_PLANE
        pfd.bReserved = 0
        pfd.dwLayerMask = 0
        pfd.dwVisibleMask = 0
        pfd.dwDamageMask = 0

        pixelFormat = ChoosePixelFormat(hdc, byref(pfd))
        SetPixelFormat(hdc, pixelFormat, byref(pfd))

        self.hrc = wglCreateContext(hdc)
        wglMakeCurrent(hdc, self.hrc)

        if version or profile:
            p = wglGetProcAddress("wglCreateContextAttribsARB".encode('utf-8'))
            if not p:
                raise Exception()
            wglCreateContextAttribsARB = WINFUNCTYPE(c_void_p, HDC, c_void_p,
                                                     POINTER(c_int))(p)

            attribs = []
            if version:
                major, minor = version.split('.')
                attribs.append(WGL_CONTEXT_MAJOR_VERSION_ARB)
                attribs.append(int(major))
                attribs.append(WGL_CONTEXT_MINOR_VERSION_ARB)
                attribs.append(int(minor))

            attribs.append(WGL_CONTEXT_FLAGS_ARB)
            attribs.append(0)

            if profile == 'core':
                attribs.append(WGL_CONTEXT_PROFILE_MASK_ARB)
                attribs.append(WGL_CONTEXT_CORE_PROFILE_BIT_ARB)
            elif profile == 'compatibility':
                attribs.append(WGL_CONTEXT_PROFILE_MASK_ARB)
                attribs.append(WGL_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB)

            attribs.append(0)

            array = (ctypes.c_int * len(attribs))(*attribs)
            hrc = wglCreateContextAttribsARB(
                hdc, None, array)
            wglDeleteContext(self.hrc)
            self.hrc = hrc
            wglMakeCurrent(hdc, self.hrc)

        for x in [
                'GL_VERSION',
                #'GL_EXTENSIONS',
                'GL_SHADING_LANGUAGE_VERSION',
                'GL_RENDERER'
        ]:
            logger.info('%s: %s', x, glGetString(globals()[x]))

        # GL_ARB_compatibility
        profile = None
        try:
            extensions = glGetString(GL_EXTENSIONS).decode('ascii').split()
            if 'GL_ARB_compatibility' in extensions:
                profile = 'COMPATIBILITY'
        except OpenGL.error.GLerror as ex:
            profile = 'CORE'
        logger.info('OpenGL%d.%d %s', glGetInteger(GL_MAJOR_VERSION),
                    glGetInteger(GL_MINOR_VERSION), profile)

    @staticmethod
    def WndProcProxy(hwnd, message, wParam, lParam):
        # sys.stderr.write("Window::WndProcProxy\n")
        p = GetWindowLongPtr(hwnd, win32con.GWL_USERDATA)
        window = cast(p, py_object)
        return window.value.WndProc(hwnd, message, wParam, lParam)


class WindowFactory(object):
    def __init__(self):
        self.classes = []
        self.windows = []
        self.wndclass = self.register_class(b"MainWin")

    def __del__(self):
        self.finalize()

    def finalize(self):
        for w in self.windows:
            w.value.finalize()
        logger.debug('finalize %s', self.__class__)

    def register_class(self, className):
        """
        RegisterClassA
        """
        # Define Window Class
        wndclass = WNDCLASS()
        wndclass.style = win32con.CS_HREDRAW | win32con.CS_VREDRAW
        wndclass.lpfnWndProc = WNDPROC(WindowFactory.WndProc)
        wndclass.cbClsExtra = wndclass.cbWndExtra = 0
        wndclass.hInstance = windll.kernel32.GetModuleHandleA(
            HANDLE(win32con.NULL))
        wndclass.hIcon = windll.user32.LoadIconA(
            HANDLE(win32con.NULL), LPCSTR(win32con.IDI_APPLICATION))
        wndclass.hCursor = windll.user32.LoadCursorA(
            HANDLE(win32con.NULL), LPCSTR(win32con.IDC_ARROW))
        # wndclass.hbrBackground = windll.gdi32.GetStockObject(
        #     c_int(win32con.WHITE_BRUSH))
        wndclass.hbrBackground = None
        wndclass.lpszMenuName = None
        wndclass.lpszClassName = className
        # Register Window Class
        res = RegisterClass(byref(wndclass))
        # logger.debug(res)
        if not res:
            raise WinError()
        self.classes.append(wndclass)
        return wndclass

    def create(self, klass, **kw):
        width = 'width' in kw and kw['width'] or win32con.CW_USEDEFAULT
        height = 'height' in kw and kw['height'] or win32con.CW_USEDEFAULT
        wndclass = b'wndclass' in kw and kw['wndclass'] or self.wndclass
        title = b'title' in kw and kw['title'] or b"glglue.wgl"

        window = klass()
        pywindow = py_object(window)
        hwnd = CreateWindowEx(0, wndclass.lpszClassName, title,
                              win32con.WS_OVERLAPPEDWINDOW,
                              win32con.CW_USEDEFAULT, win32con.CW_USEDEFAULT,
                              width, height, win32con.NULL, win32con.NULL,
                              wndclass.hInstance, pywindow)
        self.windows.append(pywindow)
        return window

    def loop(self, window):
        msg = MSG()
        pMsg = pointer(msg)
        NULL = c_int(win32con.NULL)
        # while windll.user32.GetMessageA( pMsg, NULL, 0, 0) != 0:
        #    windll.user32.TranslateMessage(pMsg)
        #    windll.user32.DispatchMessageA(pMsg)
        lastCount = timeGetTime()
        while True:
            if windll.user32.PeekMessageA(pMsg, NULL, 0, 0,
                                          win32con.PM_NOREMOVE) != 0:
                if windll.user32.GetMessageA(pMsg, NULL, 0, 0) == 0:
                    return msg.wParam
                windll.user32.TranslateMessage(pMsg)
                windll.user32.DispatchMessageA(pMsg)
            else:
                count = timeGetTime()
                d = count - lastCount
                if d > 0:
                    window.controller.onUpdate(d)
                    window.Redraw()
                    lastCount = count

    @staticmethod
    def WndProc(hwnd, message, wParam, lParam):
        # sys.stderr.write("WindowFactory::WndProc\n")
        if message == win32con.WM_CREATE:
            lpcreatestruct = cast(lParam, POINTER(CREATESTRUCT))
            createstruct = lpcreatestruct.contents
            window = cast(createstruct.lpCreateParams, py_object).value
            window.hwnd = hwnd
            window.SetLongPtr(hwnd, win32con.GWL_USERDATA,
                              createstruct.lpCreateParams)
            window.SetLongPtr(hwnd, win32con.GWL_WNDPROC,
                              cast(WNDPROC(Window.WndProcProxy), c_void_p))
            return window.WndProc(hwnd, message, wParam, lParam)

        elif message == win32con.WM_DESTROY:
            windll.user32.PostQuitMessage(0)
            return 0

        return DefWindowProc(hwnd, message, wParam, lParam)


class LoopManager:
    def __init__(self, controller, **kw):
        self.controller = controller
        self.factory = WindowFactory()
        self.window = self.factory.create(Window, **kw)
        self.window.createGLContext(16,
                                    version=kw.get('version'),
                                    profile=kw.get('profile'))
        self.window.controller = controller
        ShowWindow(self.window.hwnd, win32con.SW_SHOWNORMAL)
        self.msg = MSG()
        self.pMsg = pointer(self.msg)
        self.NULL = c_int(win32con.NULL)

    def begin_frame(self):
        while True:
            if windll.user32.PeekMessageA(self.pMsg, self.NULL, 0, 0,
                                          win32con.PM_NOREMOVE) != 0:
                if windll.user32.GetMessageA(self.pMsg, self.NULL, 0, 0) == 0:
                    return  #msg.wParam
                windll.user32.TranslateMessage(self.pMsg)
                windll.user32.DispatchMessageA(self.pMsg)
            break

        return timeGetTime()

    def end_frame(self):
        ps = PAINTSTRUCT()
        hdc = BeginPaint(self.window.hwnd, byref(ps))
        if hdc:
            SwapBuffers(hdc)
        EndPaint(self.window.hwnd, byref(ps))


def mainloop(controller, **kw):
    loop = LoopManager(controller, **kw)
    lastCount = None
    while True:
        count = loop.begin_frame()
        if not count:
            break
        d = count - lastCount if lastCount else 0
        lastCount = count
        if d > 0:
            controller.onUpdate(d)
            controller.draw()
            loop.end_frame()


if __name__ == "__main__":
    from logging import basicConfig, DEBUG
    basicConfig(format='%(levelname)s:%(name)s:%(message)s', level=DEBUG)
    import os
    if os.name != 'nt':
        logger.debug("this script is windows only: " + os.name)
        sys.exit()
    import glglue.sample
    mainloop(glglue.sample.SampleController(),
             width=600,
             height=400,
             version="3.2")



import os
import sys
import unittest
import pathlib
import glglue.gltf


GLTF_SAMPLE_DMODELS = pathlib.Path(os.environ['GLTF_SAMPLE_MODELS'])
if not GLTF_SAMPLE_DMODELS.exists():
    sys.exit()


class TestGltf(unittest.TestCase):

    def test_gltf(self):
        path = GLTF_SAMPLE_DMODELS / '2.0/Box/glTF/Box.gltf'
        gltf_data = glglue.gltf.parse_path(path)
        self.assertEqual(1, len(gltf_data.materials))
        self.assertEqual(1, len(gltf_data.meshes))
        self.assertEqual(
            24, gltf_data.meshes[0].primitives[0].position.count())
        self.assertEqual(24, gltf_data.meshes[0].primitives[0].normal.count())
        self.assertEqual(36, gltf_data.meshes[0].primitives[0].indices.count())
        self.assertEqual(
            'Red', gltf_data.meshes[0].primitives[0].material.name)
        self.assertEqual(2, len(gltf_data.nodes))

    def test_glb(self):
        path = GLTF_SAMPLE_DMODELS / '2.0/BoxTextured/glTF-Binary/BoxTextured.glb'
        gltf_data = glglue.gltf.parse_path(path)
        self.assertEqual(1, len(gltf_data.materials))
        self.assertEqual(1, len(gltf_data.meshes))
        self.assertEqual(
            24, gltf_data.meshes[0].primitives[0].position.count())
        self.assertEqual(24, gltf_data.meshes[0].primitives[0].normal.count())
        self.assertEqual(24, gltf_data.meshes[0].primitives[0].uv.count())
        self.assertEqual(36, gltf_data.meshes[0].primitives[0].indices.count())
        self.assertEqual(
            'Texture', gltf_data.meshes[0].primitives[0].material.name)
        self.assertEqual('CesiumLogoFlat.png',
                         gltf_data.textures[0].image.name)
        self.assertEqual(glglue.gltf.MimeType.Png,
                         gltf_data.textures[0].image.mime)
        self.assertEqual(2, len(gltf_data.nodes))


if __name__ == '__main__':
    unittest.main()



from PySide6 import QtWidgets, QtGui, QtCore
import pathlib
import logging
import glglue.gltf
from glglue.scene.texture import Texture, Image32
from glglue.scene.material import Material
from glglue.scene.mesh import Mesh
from glglue.scene.node import Node
from glglue.scene.vertices import Planar
import glglue.gl3.vbo
import glglue.gl3.shader
from typing import Dict, List
from OpenGL import GL

logger = logging.getLogger(__name__)
logging.basicConfig(format='%(levelname)s:%(name)s:%(message)s',
                    level=logging.DEBUG)


VS = '''
in vec3 aPosition;

#ifdef HAS_UV
in vec2 aUV;
out vec2 vUV;
#endif

uniform mediump mat4 m;
uniform mediump mat4 vp;

void main ()
{
    gl_Position = vec4(aPosition, 1) * m * vp;
#ifdef HAS_UV
    vUV = aUV;
#endif
}
'''

FS = '''
#ifdef HAS_UV
in vec2 vUV;
#else
vec2 vUV = vec2(1, 1);
#endif

uniform sampler2D COLOR_TEXTURE;

out vec4 fColor;
void main()
{
    vec4 tex = texture(COLOR_TEXTURE, vUV);
    fColor = tex;
}
'''


class Loader:
    def __init__(self, gltf: glglue.gltf.GltfData) -> None:
        self.gltf = gltf
        self.images: Dict[glglue.gltf.GltfImage, Image32] = {}
        self.textures: Dict[glglue.gltf.GltfTexture, Texture] = {}
        self.materials: Dict[glglue.gltf.GltfMaterial, Material] = {}
        self.meshes: Dict[glglue.gltf.GltfPrimitive, Mesh] = {}

    def _load_image(self, src: glglue.gltf.GltfImage):
        image = Image32.load(src.data)
        self.images[src] = image

    def _load_texture(self, src: glglue.gltf.GltfTexture):
        texture = Texture(src.name, self.images[src.image])
        self.textures[src] = texture

    def _load_material(self, src: glglue.gltf.GltfMaterial):
        material = Material(src.name, VS, FS)
        if src.base_color_texture:
            material.color_texture = self.textures[src.base_color_texture]
        self.materials[src] = material

    def _load_mesh(self, name: str, src: glglue.gltf.GltfPrimitive):
        macro = ['#version 330']
        attributes: List[glglue.gl3.vbo.TypedBytes] = [
            glglue.gl3.vbo.TypedBytes(*src.position)]
        # if prim.normal:
        #     attributes.append(glglue.gl3.vbo.TypedBytes(*prim.normal))
        #     macro += f'#define HAS_NORMAL 1\n'
        if src.uv:
            attributes.append(glglue.gl3.vbo.TypedBytes(*src.uv))
            macro.append('#define HAS_UV 1')
        indices = None
        if src.indices:
            indices = glglue.gl3.vbo.TypedBytes(*src.indices)

        mesh = Mesh(name, Planar(attributes), indices)
        mesh.add_submesh(self.materials[src.material], macro, GL.GL_TRIANGLES)
        self.meshes[src] = mesh

    def _load(self, src: List[glglue.gltf.GltfNode], dst: Node):
        for gltf_node in src:
            node = Node(gltf_node.name)
            dst.children.append(node)

            if gltf_node.mesh:
                for gltf_prim in gltf_node.mesh.primitives:
                    mesh = self.meshes[gltf_prim]
                    node.meshes.append(mesh)

            self._load(gltf_node.children, node)

    def load(self) -> Node:
        for image in self.gltf.images:
            self._load_image(image)
        for texture in self.gltf.textures:
            self._load_texture(texture)
        for material in self.gltf.materials:
            self._load_material(material)
        for mesh in self.gltf.meshes:
            for i, prim in enumerate(mesh.primitives):
                self._load_mesh(f'{mesh.name}:{i}', prim)

        scene = Node('__scene__')
        self._load(self.gltf.scene, scene)
        return scene


class Window(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent)
        # setup opengl widget
        import glglue.gl3.samplecontroller
        self.controller = glglue.gl3.samplecontroller.SampleController()
        import glglue.pyside6
        import glglue.utils
        self.glwidget = glglue.pyside6.Widget(
            self, self.controller, glglue.utils.get_desktop_scaling_factor())
        self.setCentralWidget(self.glwidget)

        # menu
        menu = self.menuBar()
        file_menu = menu.addMenu("&File")
        open_action = QtGui.QAction("&Open", self)
        open_action.triggered.connect(self.open_dialog)  # type: ignore
        file_menu.addAction(open_action)

    @QtCore.Slot()  # type: ignore
    def open_dialog(self):
        dialog = QtWidgets.QFileDialog(self, caption="open bvh")
        dialog.setFileMode(QtWidgets.QFileDialog.AnyFile)
        dialog.setNameFilters(
            ["gltf files (*.gltf;*.glb;*.zip;*.vrm)", "Any files (*)"])
        if not dialog.exec():
            return
        files = dialog.selectedFiles()
        if not files:
            return
        path = pathlib.Path(files[0])
        self.load(path)

    def load(self, path: pathlib.Path):
        gltf = glglue.gltf.parse_path(path)

        # create mesh
        loader = Loader(gltf)
        scene = loader.load()
        self.controller.drawables = [scene]
        self.glwidget.repaint()


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    window = Window()
    window.resize(1600, 1200)

    if len(sys.argv) > 1:
        path = pathlib.Path(sys.argv[1])
        window.load(path)

    window.show()
    sys.exit(app.exec())



'''
https://github.com/KhronosGroup/glTF/blob/main/specification/2.0/
'''
from ctypes.wintypes import RGB
from typing import NamedTuple, Optional, Tuple, List, Dict, Any, Type
import struct
import ctypes
from enum import Enum
import json
import pathlib
from dataclasses import dataclass


class GltfError(RuntimeError):
    pass


COMPONENT_TYPE_TO_ELEMENT_TYPE = {
    5120: ctypes.c_char,
    5121: ctypes.c_byte,
    5122: ctypes.c_short,
    5123: ctypes.c_ushort,
    5125: ctypes.c_uint,
    5126: ctypes.c_float,
}

TYPE_TO_ELEMENT_COUNT = {
    'SCALAR': 1,
    'VEC2': 2,
    'VEC3': 3,
    'VEC4': 4,
    'MAT2': 4,
    'MAT3': 9,
    'MAT4': 16,
}


class MimeType(Enum):
    Jpg = "image/jpeg"
    Png = "image/png"

    @staticmethod
    def from_name(name: str):
        match pathlib.Path(name).suffix.lower():
            case ".png":
                return MimeType.Png
            case ".jpg":
                return MimeType.Jpg
            case _:
                raise GltfError(f'unknown image: {name}')


def get_accessor_type(gltf_accessor) -> Tuple[Type[ctypes._SimpleCData], int]:
    return COMPONENT_TYPE_TO_ELEMENT_TYPE[gltf_accessor['componentType']], TYPE_TO_ELEMENT_COUNT[gltf_accessor['type']]


class TypedBytes(NamedTuple):
    data: bytes
    element_type: Type[ctypes._SimpleCData]
    element_count: int = 1

    def stride(self) -> int:
        return ctypes.sizeof(self.element_type) * self.element_count

    def count(self) -> int:
        return len(self.data) // self.stride()


class GltfBufferReader:
    def __init__(self, gltf, path: Optional[pathlib.Path], bin: Optional[bytes]):
        self.gltf = gltf
        self.bin = bin
        self.path = path
        self.uri_cache: Dict[str, bytes] = {}

    def uri_bytes(self, uri: str) -> bytes:
        data = self.uri_cache.get(uri)
        if not data:
            if self.path:
                path = self.path.parent / uri
                data = path.read_bytes()
            else:
                raise NotImplementedError()
        self.uri_cache[uri] = data
        return data

    def _buffer_bytes(self, buffer_index: int) -> bytes:
        if self.bin and buffer_index == 0:
            # glb bin_chunk
            return self.bin

        gltf_buffer = self.gltf['buffers'][buffer_index]
        uri = gltf_buffer['uri']
        if not isinstance(uri, str):
            raise GltfError()

        return self.uri_bytes(uri)

    def buffer_view_bytes(self, buffer_view_index: int) -> bytes:
        gltf_buffer_view = self.gltf['bufferViews'][buffer_view_index]
        bin = self._buffer_bytes(gltf_buffer_view['buffer'])
        offset = gltf_buffer_view.get('byteOffset', 0)
        length = gltf_buffer_view['byteLength']
        return bin[offset:offset+length]

    def read_accessor(self, accessor_index: int) -> TypedBytes:
        gltf_accessor = self.gltf['accessors'][accessor_index]
        bin = self.buffer_view_bytes(gltf_accessor['bufferView'])
        offset = gltf_accessor.get('byteOffset', 0)
        count = gltf_accessor['count']
        element_type, element_count = get_accessor_type(gltf_accessor)
        bin = bin[offset:offset+ctypes.sizeof(element_type) *
                  element_count*count]
        return TypedBytes(bin, element_type, element_count)


class GltfImage(NamedTuple):
    name: str
    data: bytes
    mime: MimeType


class GltfTexture(NamedTuple):
    name: str
    image: GltfImage


class RGBA(NamedTuple):
    r: float
    g: float
    b: float
    a: float


class GltfMaterial(NamedTuple):
    name: str
    base_color_texture: Optional[GltfTexture]
    base_color_factor: RGBA
    metallic_factor: float


class GltfPrimitive(NamedTuple):
    material: GltfMaterial
    position: TypedBytes
    normal: Optional[TypedBytes]
    uv: Optional[TypedBytes]
    indices: Optional[TypedBytes]


class GltfMesh(NamedTuple):
    name: str
    primitives: Tuple[GltfPrimitive, ...]


@dataclass
class GltfNode:
    name: str
    children: List['GltfNode']
    mesh: Optional[GltfMesh]


class GltfData:
    def __init__(self, gltf):
        self.gltf = gltf
        self.images: List[GltfImage] = []
        self.textures: List[GltfTexture] = []
        self.materials: List[GltfMaterial] = []
        self.meshes: List[GltfMesh] = []
        self.nodes: List[GltfNode] = []
        self.scene: List[GltfNode] = []

    def __str__(self) -> str:
        return f'{len(self.materials)} materials, {len(self.meshes)} meshes, {len(self.nodes)} nodes'

    def _parse_image(self, buffer_reader: GltfBufferReader, i: int, gltf_image) -> GltfImage:
        name = gltf_image.get('name')
        match gltf_image:
            case {'bufferView': buffer_view_index, 'mimeType': mime}:
                return GltfImage(name or f'{i}', buffer_reader.buffer_view_bytes(buffer_view_index), MimeType(mime))
            case {'uri': uri}:
                if uri.startswith('data:'):
                    raise NotImplementedError()
                else:
                    return GltfImage(uri, buffer_reader.uri_bytes(uri), MimeType.from_name(uri))
            case _:
                raise GltfError()

    def _parse_texture(self, i: int, gltf_texture) -> GltfTexture:
        texture = GltfTexture(gltf_texture.get(
            'name', f'{i}'), self.images[gltf_texture['source']])
        return texture

    def _parse_material(self, i: int, gltf_material) -> GltfMaterial:
        pbr = gltf_material.get('pbrMetallicRoughness')
        base_color_texture = None
        base_color_factor = RGBA(1, 1, 1, 1)
        metallic_factor = 0.0
        if pbr:
            match pbr.get('baseColorTexture'):
                case {'index': texture_index}:
                    base_color_texture = self.textures[texture_index]
            base_color_factor = RGBA(*pbr.get(
                'baseColorFactor', [1, 1, 1, 1]))
            metallic_factor = pbr.get('metallicFactor', 0.0)

        material = GltfMaterial(gltf_material.get(
            'name', f'{i}'), base_color_texture, base_color_factor, metallic_factor)
        return material

    def _parse_mesh(self, buffer_reader: GltfBufferReader, i: int, gltf_mesh) -> GltfMesh:
        primitives = []
        for gltf_prim in gltf_mesh['primitives']:
            gltf_attributes = gltf_prim['attributes']

            match gltf_attributes['POSITION']:
                case int() as accessor:
                    positions = buffer_reader.read_accessor(accessor)
                case _:
                    raise GltfError()

            normal = None
            match gltf_attributes.get('NORMAL'):
                case int() as accessor:
                    normal = buffer_reader.read_accessor(accessor)

            uv = None
            match gltf_attributes.get('TEXCOORD_0'):
                case int() as accessor:
                    uv = buffer_reader.read_accessor(accessor)

            indices = None
            match gltf_prim.get('indices'):
                case int() as accessor:
                    indices = buffer_reader.read_accessor(accessor)

            prim = GltfPrimitive(
                self.materials[gltf_prim['material']], positions, normal, uv, indices)
            primitives.append(prim)

        mesh = GltfMesh(gltf_mesh.get('name', f'{i}'), tuple(primitives))
        return mesh

    def _parse_node(self, i: int, gltf_node) -> GltfNode:
        node = GltfNode(gltf_node.get('name', f'{i}'), [], None)
        match gltf_node.get('mesh'):
            case int() as mesh_index:
                node.mesh = self.meshes[mesh_index]
        return node

    def parse(self, buffer_reader: GltfBufferReader):
        # image
        for i, gltf_image in enumerate(self.gltf.get('images', [])):
            image = self._parse_image(buffer_reader, i, gltf_image)
            self.images.append(image)

        # texture
        for i, gltf_texture in enumerate(self.gltf.get('textures', [])):
            texture = self._parse_texture(i, gltf_texture)
            self.textures.append(texture)

        # material
        maetrials = self.gltf.get('materials')
        if maetrials:
            for i, gltf_material in enumerate(maetrials):
                material = self._parse_material(i, gltf_material)
                self.materials.append(material)
        else:
            NotImplementedError('no material')

        # mesh
        for i, gltf_mesh in enumerate(self.gltf.get('meshes', [])):
            mesh = self._parse_mesh(buffer_reader, i, gltf_mesh)
            self.meshes.append(mesh)

        # node
        for i, gltf_node in enumerate(self.gltf.get('nodes', [])):
            node = self._parse_node(i, gltf_node)
            self.nodes.append(node)
        for i, gltf_node in enumerate(self.gltf.get('nodes', [])):
            match gltf_node.get('children'):
                case [*children]:
                    for child_index in children:
                        self.nodes[i].children.append(self.nodes[child_index])

        # skinning
        for i, gltf_skin in enumerate(self.gltf.get('skins', [])):
            pass

        # scene
        self.scene += [self.nodes[node_index]
                       for node_index in self.gltf['scenes'][self.gltf['scene']]['nodes']]


def parse_gltf(json_chunk: bytes, *, path: Optional[pathlib.Path] = None, bin: Optional[bytes] = None) -> GltfData:
    gltf = json.loads(json_chunk)
    data = GltfData(gltf)
    buffer_reader = GltfBufferReader(gltf, path, bin)
    data.parse(buffer_reader)

    return data


class GlbError(RuntimeError):
    pass


class BytesReader:
    def __init__(self, data: bytes):
        self.data = data
        self.pos = 0

    def is_end(self) -> bool:
        return self.pos >= len(self.data)

    def read(self, size: int) -> bytes:
        if (self.pos + size) > len(self.data):
            raise IOError()
        data = self.data[self.pos:self.pos+size]
        self.pos += size
        return data

    def read_int(self) -> int:
        data = self.read(4)
        return struct.unpack('i', data)[0]


def parse_glb(data: bytes) -> Tuple[Optional[bytes], Optional[bytes]]:
    '''
    https://www.khronos.org/registry/glTF/specs/2.0/glTF-2.0.html#glb-file-format-specification
    '''
    r = BytesReader(data)
    if r.read_int() != 0x46546C67:
        raise GlbError('invalid magic')

    version = r.read_int()
    if version != 2:
        raise GlbError(f'unknown version: {version}')

    length = r.read_int()

    json_chunk = None
    bin_chunk = None
    while r.pos < length:
        chunk_length = r.read_int()
        chunk_type = r.read_int()
        chunk_data = r.read(chunk_length)
        match chunk_type:
            case 0x4E4F534A:
                json_chunk = chunk_data
            case 0x004E4942:
                bin_chunk = chunk_data
            case _:
                raise NotImplementedError(f'unknown chunk: {chunk_type}')

    return json_chunk, bin_chunk


def parse_path(path: pathlib.Path) -> GltfData:
    data = path.read_bytes()
    try:
        # first, try glb
        json, bin = parse_glb(data)
        if json and bin:
            return parse_gltf(json, path=path, bin=bin)
    except GlbError:
        pass

    # fallback to gltf
    return parse_gltf(data, path=path)



# coding: utf-8
from logging import getLogger

logger = getLogger(__name__)

from logging import basicConfig, DEBUG

basicConfig(format='%(levelname)s:%(name)s:%(message)s', level=DEBUG)

if __name__ == "__main__":
    from glglue.sample.imguicontroller import ImGuiController
    controller = ImGuiController()

    import glglue.wgl
    loop = glglue.wgl.LoopManager(controller,
                                  width=640,
                                  height=480,
                                  title=b"imgui")

    lastCount = 0
    while True:
        count = loop.begin_frame()
        if not count:
            break
        d = count - lastCount
        lastCount = count
        if d > 0:
            controller.onUpdate(d)
            controller.draw()
            loop.end_frame()
